#include <set>
#include <map>
#include "cverilog.hpp"

namespace CVER {

  // Cverilog main class

  Cverilog *current;
  wire *w_0, *w_1, *w_finished;
  int monitor_level;
  
  Cverilog::Cverilog(ostream &o, ostream &odot): o(o), odot(odot) {
      current=this;
      w_0=new wire("w_0"); 
      w_1=new wire("w_1");
      tick=new wire("tick");
      rst=new wire("rst");
      w_0->monitor=w_1->monitor=rst->monitor=0;
      wires.clear();
      w_finished=new wire("w_finished");
      w_finished -> output = w_finished->required = 1;
      w_finished->monitor=3;

      w_prefix="w_";
      labwire=2;
      labcirc=0;
      monitor_level=1;
      delay=0; // default
 }
    

  void Cverilog::build()
  {

    int nc,hc;

    // move w_finished to end, for display purposes
    if(wires[0]==w_finished){
      wires.erase(wires.begin());
      wires.push_back(w_finished);
    }

    sr_ff::pr_def();
    d_ff::pr_def();
    counter::pr_def();

    o<<"module testbench;\n";
    for(vector<wire*>::iterator w=wires.begin(); w!=wires.end(); w++)
      o<<"wire "<<(*w)->name<<";\n";

    o<<"reg w_0;\n";
    o<<"reg w_1;\n";
    o<<"reg tick;\n";
    o<<"reg rst;\n\n";
    //    o<<"wire w_finished;\n";

    for(vector<vermod*>::iterator m=modules.begin(); m!=modules.end();m++)
      (*m)->pr_instance();

    // simulation
       
    o<<"initial begin\n";

    nc=0; hc=1;

    while(hc){
      hc=0;
      if(!nc) o<<"  $display(\"                time "; else
	o<<      "  $display(\"                     ";
      for(vector<wire*>::iterator w=wires.begin(); w!=wires.end(); w++)
	if((*w)->monitor>=monitor_level)
	  if((int)(*w)->name.size()>nc){
	    o<<" "<<(*w)->name[nc];
	    hc=1;
	  }
	  else o<<"  ";
      o<<"\");\n";
      nc++;
    }

    o<<"  $monitor($time,";
    for(vector<wire*>::iterator w=wires.begin(); w!=wires.end(); w++)
      if((*w)->monitor>=monitor_level)
	o<<",\" \","<<(*w)->name;
    o<<");\n";

    o<<"  w_0=0; w_1=1;\n";
    o<<"  tick=0; rst=1;\n  #500 rst=0;\n";
    //    o<<"  while($time<100) begin\n";
    //    o<<"  while($time < 1000000010 && w_finished == 0) begin\n";
    o<<"  while(w_finished == 0) begin";
    o<<"     #500 tick=1;\n";
    o<<"     #500 tick=0;\n";
    o<<"  end\n";
    o<<"end\n";

    o<<"endmodule\n";
  }

  //this will be called by the user program for debugging
  string Cverilog::wn(wire *w)
  {
    if(w) return w->name;
    return "NULL";
  }

  // this is called by modules in pr_instance()
  string vermod::wn(wire *w, int def)
  {
    if(!w) return def ? "1" : "0";
    if(w==w_0) return "0";
    if(w==w_1) return "1";
    return w->name;
  }

  void vermod::drawin(wire *w, string edge_label)
  {
    if(w)(current->odot)<<w->name<<" -> "<<name<<"[ label=\""<<edge_label<<"\"]\n";
  }

  void vermod::drawout(wire *w, string edge_label)
  {
    if(w)(current->odot)<<name<<" -> "<< w->name<<"[ label=\""<<edge_label<<"\"]"<<'\n';
  }

  void vermod::drawin(vector<wire *> &w)
  {
    for (vector<wire *>::iterator i = w.begin(); i != w.end(); i++)
      drawin((*i),"");
  }

  void Cverilog::drawdot()
  {
    odot<<"digraph circuit {\nsize=\"7.5,10\"\n";

    w_0->dw_dot();
    w_1->dw_dot();
    //    w_finished->dw_dot();

    for(vector<wire*>::iterator w=wires.begin(); w!=wires.end(); w++){
      (*w)->dw_dot();
      }
    
    for(vector<vermod*>::iterator m=modules.begin(); m!=modules.end();m++){
      //      std::cerr<<"in\n";
      if(*m) {
	//std::cerr<<"name="<<(*m)->name<<'\n';
	(*m)->dw_dot();
      } else assert(0);
      //std::cerr<<"out\n";
    }

    odot<<"}\n";
  }

  wire* Cverilog::build_orgate(vector<wire*> &win)
  {

    int sz;
    orgate *ogate;
    wire *w;
    
    // NULL and w_0 ignored / w_1 shortcuts ; no wires -> 0


    vector<wire*>::iterator iv = win.begin();
    while(iv != win.end()){
      if((*iv) == w_1) return w_1;
      if((*iv) == NULL || (*iv) == w_0) iv = win.erase(iv); else iv++;
    }

    sort(win.begin(), win.end());
    win.erase(unique(win.begin(), win.end()), win.end());

    sz = win.size();
    if(sz == 0) return NULL;
    if(sz == 1) return win[0];

    for(vector<vermod*>::iterator m = current->modules.begin(); m != current->modules.end(); m++){
      ogate = dynamic_cast<orgate*> (*m);
      if(!ogate) continue;
      if(ogate->win == win) return ogate->out;
    }

    ogate = new orgate();
    ogate->win = win;
    w=new wire(); ogate->out=w;
    return w;
  }

  wire* Cverilog::build_orgate(wire *w1, wire *w2)
  {
    vector<wire*> tmp;
    tmp.push_back(w1);
    tmp.push_back(w2);
    return build_orgate(tmp);
  }

  wire* Cverilog::build_andgate(vector<wire*> &win)
  {
    int sz;
    andgate *agate;
    wire *w;
    
    // w_1 ignored / NULL or w_0 shortcuts  ; no wires -> 1

    vector<wire*>::iterator iv = win.begin();
    while(iv != win.end()){
      if((*iv) == NULL || (*iv)==w_0) return NULL;
      if((*iv) == w_1) iv = win.erase(iv); else iv++;
    }

    sort(win.begin(), win.end());
    win.erase(unique(win.begin(), win.end()), win.end());

    sz = win.size();
    if(sz == 0) return w_1;
    if(sz == 1) return win[0];

    for(vector<vermod*>::iterator m = current->modules.begin(); m != current->modules.end(); m++){
      agate = dynamic_cast<andgate*> (*m);
      if(!agate) continue;

      if(agate->win == win) return agate->out;
    }

    agate = new andgate();
    agate->win = win;
    w=new wire(); agate->out=w;
    return w;
   }

  wire* Cverilog::build_andgate(wire *w1, wire *w2)
  {
    vector<wire*> tmp;
    tmp.push_back(w1);
    tmp.push_back(w2);
    return build_andgate(tmp);
  }

  wire* Cverilog::build_notgate(wire *w)
  {
    notgate *ng;
    
    wire *wn;
    if(w==NULL || w==w_0) return w_1;
    if(w==w_1) return NULL;


    for(vector<vermod*>::iterator m = current->modules.begin(); m != current->modules.end(); m++){
      ng = dynamic_cast<notgate*> (*m);
      if(!ng) continue;

      if( w == ng->in ) return ng->out;
      if( w == ng->out) return ng->in;

    }


    wn=new wire();
    new notgate(wn,w);
    return wn;
  }



  // REMOVE UNUSED COMPONENTS

  void Cverilog::remove_unused()
  {
    vector<vermod*>::iterator im;

    vector<wire*> win, wout;
    vector<wire*>::iterator iw, jw;

    map<wire*, vermod*> wowner;
    map<wire*, vector<wire*> > wneeded;
    wire *w;

    vector<wire*> wreq;


    for(im=modules.begin();im!=modules.end();im++){

      win.clear(); wout.clear();
      (*im)->req_wire(win, wout);

      assert(wout.size()==1); //temporary
      w = wout[0];

      wowner[w] = *im;
      wneeded[w]= win;

    }

    //    wreq.push_back(w_finished); // required !
    for(iw=wires.begin(); iw!=wires.end();iw++){
      if((*iw)->output) (*iw)->required = 1; // on output is required !
      if((*iw)->required) {
	wreq.push_back(*iw);
      }
    }

    // checks wire dependency

    modules.clear();
    iw=wreq.begin();
    while(iw != wreq.end()){

      w=*iw; wreq.erase(iw);

      if(wowner.count(w) > 0){
	modules.push_back(wowner[w]);

	for(jw=wneeded[w].begin(); jw!=wneeded[w].end(); jw++){
	  if((*jw)->required) continue; // already processed
	  wreq.push_back(*jw);
	  (*jw)->required=10;
	}
      }
      
      iw=wreq.begin();
    }
    
    // removes unused wires and restore the require level
    wreq.clear();
    for(iw=wires.begin(); iw!=wires.end(); iw++){
      if((*iw)->required) wreq.push_back(*iw);
      if((*iw)->required==10) (*iw)->required=0;
    }
    wires = wreq;

  }




  // WIRE

  wire::wire(string name): name(name), input(0), output(0){
    current->wires.push_back(this);
    monitor=1;
    required=0;
  }

  wire::wire(string name, int i): input(0), output(0)
  {
    stringstream s;
    s<<name<<i;
    this->name=s.str(); 
    current->wires.push_back(this);
    monitor=1;
    required=0;
  }

  wire::wire(): input(0), output(0) {
    name=current->new_wname(); current->wires.push_back(this);
    monitor=0;
    required=0;
  }

  void wire::dw_dot()
  {
    string label;

    label = (input || output || monitor >= monitor_level) ? name : "";
    if(this==w_0) label="0"; if(this==w_1) label="1"; if(this==w_finished) label="STOP";


    current->odot<<name<<" [ label=\""<<label<<"\" fixed_size=true height=\"0.2\" width=\"0.2\" ";
    current->odot <<" fontsize=\"10\" ]\n";
  }



  // STRAP

  strap::strap(wire *out, wire *in): out (out), in(in)
  {
    assert(out);
    assert(in);
    name = current->new_wname();
    current->modules.push_back(this);
  }

  strap::strap(wire *probe, string wname, int monitor)
  {
    wire *w;

    if(!probe) probe=w_0;

    w=new wire(wname);
    w->monitor = monitor;
    out = w; in = probe;

    name = current ->new_wname();
    current->modules.push_back(this);
  }
  
  strap::strap(wire *probe, string wname, int ix, int monitor)
  {
    wire *w;
    stringstream w_name;

    if(!probe) probe=w_0;

    w_name << wname << ix;
    w=new wire(w_name.str());
    w->monitor = monitor;
    out = w; in = probe;

    name = current ->new_wname();
    current->modules.push_back(this);
  }

  void strap::pr_instance()
  {
    ostream &o=current->o;
    assert(out&&in);
    o<<"assign "<<out->name<<"="<<in->name<<";\n";
  }

  void strap::dw_dot()
  {
    assert(out && in);
    ostream &o=current->odot;
    o<<in->name<<" -> "<<out->name<<'\n';
  }

  void strap::req_wire(vector<wire*> &win, vector<wire*> &wout)
  {
    assert(in && out);
    win.push_back(in);
    wout.push_back(out);
  }

}
