#include "grc2zaphod.hpp"

using namespace AST;
namespace ZAPHOD {

  static char *type_names[6] = {"boot","leaf","final","pass","cache","ornet"};

  int debug, sdebug, edebug, spdebug;

  void GRC2zaphod::print_wirevec(string comment, vector<wire*> &wv)
  {
    std::cerr<<comment;
    for(vector<wire*>::iterator w = wv.begin(); w!=wv.end(); w++)
      std::cerr<<" "<<current->wn(*w);
    std::cerr<<'\n';
  }

  statemachine::statemachine(STNode *n, GRC2zaphod *z):
    owner(n), zaphod(z)
  {
    GRCNode *c;
    STleaf *lf;
    STref *ref;
    int i;
    vector<wire*> ornet;
    int activeok;

    if(debug)std::cerr<<"new machine for "<<zaphod->refmap[n]<<"\n";

    id=zaphod->refmap[n];
    nostates=n->successors.size();
    fststate = nostates - 1;

    checks.resize(nostates);
    enters.resize(nostates); enter_w.resize(nostates);
    leaves.resize(nostates); leave_w.resize(nostates);
    ff_type.resize(nostates);
    ornet.resize(nostates);

    for(i=0; i<nostates; i++){
      enters[i] = leaves[i] = NULL;

      c=n->successors[i];
      assert(c);

      if(dynamic_cast<STpar*>(c)){
	ff_type[i] = typepass;
	checks[i] = w_1; ornet[i]=NULL; continue;
      }

      if((lf=dynamic_cast<STleaf*>(c))) {
	checks[i] = new wire("q", zaphod->refmap[n]*100+i);
	checks[i]->monitor = 2;
	ornet[i] = NULL; // leaf ornet disabled ***
	if( lf == zaphod->boot) {ff_type[i] = typeboot; continue;}
	if( lf -> isfinal()) {ff_type[i]=typefin; continue;}
	ff_type[i] = typeleaf; continue;
      }
      
      if((ref=dynamic_cast<STref*>(c))){
	STexcl *below; wire *w;
	c = n->successors[i] ->successors[0];
	below = dynamic_cast<STexcl*>(c);
	if(below && 0){ // *** ref ornet disabled
	  w = zaphod->machines[below]->getactive();
	} else w = NULL;
	ornet[i] = w;

	if(ref->issuspend()){ ff_type[i] = typepass; checks[i]=w_1; continue;}
	if(ref->isabort()){
	  if(w) {ff_type[i] = typeornet; checks[i] = w; continue;}
	  ff_type[i] = typecache; 
	  checks[i]= new wire("q", zaphod->refmap[n]*100+i);
	  checks[i]->monitor = 2;
	  ornet[i] = checks[i];
	  continue;
	}
	assert(0);
      }	  
      assert(0);
    }

    // compute the active wire, if it's not too complicated
    activeok=1;
    for(i=0;i<nostates;i++)
      if(ornet[i] == NULL) {activeok = 0; break;}
    if(nostates > 2) activeok=0;

    active = activeok ? current->build_orgate(ornet) : NULL;
    
  }

  // STATE MACHINE

  void statemachine::addenter(int no, wire *w, Sync *sy)
  {
    if(sdebug)std::cerr<<"... addenter "<<no<<" to sm "<<id<<" : "<< current->wn(w) <<
		" with trap context "<<zaphod->refmap[sy]<<'\n';
    enter_w[no][sy].push_back(w);
  }

  wire* statemachine::getstate(int no)
  {
    return checks[no];
  }

  wire* statemachine::getactive() { return active; }

  // WIRE FORM TRAPMAP - builds the actual control signals for flip flops

  wire* statemachine::wire_from_trapmap(trapmap &tm, int fftype)
  {
    vector<wire*> all_levels;
    trapmap::iterator itm;
    wire *w, *wevalid;
    Sync *sy;

    if(fftype == typepass || fftype == typeornet) return NULL;

    for(itm=tm.begin(); itm!=tm.end(); itm++){
      w=current->build_orgate((*itm).second); // or of enters on the level
            if(sdebug) zaphod->print_wirevec("          enters: ",(*itm).second);
      sy = (*itm).first;
      if(fftype == typefin)
	wevalid = zaphod->fin_loop_valid[sy];
      else
	wevalid = zaphod->local_enters_valid[sy];
      
            if(sdebug)std::cerr<<"    level wire "<<current->wn(w)<<'\n';
            if(sdebug)std::cerr<<"    wevalid "<<current->wn(wevalid)<<'\n';
      w=current->build_andgate(w, wevalid);
            if(sdebug)std::cerr<<"    and wire "<<current->wn(w)<<'\n';
      all_levels.push_back(w);
    }
    w = current->build_orgate(all_levels);
        if(sdebug)std::cerr<<"  total or wire "<<current->wn(w)<<'\n';
    return w;

  }


  // dont's care about threads which do not t0, we do it later
  void statemachine::sm2circuit(wire *susp)
  {
    d_ff* dff;
    int i, ix;
    Sync *sy;

    assert(zaphod->next_sync.count(owner));
    sy = zaphod->next_sync[owner];
    if(debug)std::cerr<<"***** Building machine for "<<zaphod->refmap[owner]<<"    fststate="<<fststate<<" ******\n";

    for(i=0; i<nostates; i++){
      if(sdebug) std::cerr<<"state #"<<i<<'\n';
      if(spdebug)std::cerr<<"   ff["<<i<<"] adding suspend self enter "<<current->wn(susp)
			  <<" w/ sync context "<<zaphod->refmap[sy]<<"\n";
      enter_w[i][sy].push_back(current->build_andgate(checks[i], susp));
      enters[i] = wire_from_trapmap(enter_w[i], ff_type[i]);
    }

    if(sdebug)std::cerr<<"*** Building flip flops ...\n";
    for(i=0; i<nostates; i++) {
      ix = zaphod->refmap[owner]*100 + i;
      if(sdebug) std::cerr<<" ff "<<i<<" is of type"<< type_names [ff_type[i]]<<'\n';
      switch(ff_type[i]){
      case typeboot:
	dff = new d_ff(checks[i], w_0, "D", ix);
	dff -> init = 1;
	break;
      case typeleaf:
      case typefin:
	dff = new d_ff(checks[i], enters[i], "D",ix);
	break;
      case typecache:
	if(sdebug) cerr<<"... building cache ff\n";
	dff = new d_ff(checks[i], enters[i], "C", ix);
	break;
      case typepass:
	break;
      case typeornet:
	if(sdebug) cerr<<"... using ornet\n";
	break;
      }
    }
  }

  void statemachine::dump() {}


  // GRC2zaphod
  
  void GRC2zaphod::init()
  {
    stroot=dynamic_cast<STNode*>(top->successors[0]);
    assert(stroot);
    boot=dynamic_cast<STleaf*>(stroot->successors.back());
    assert(boot);

    top_sw=dynamic_cast<Switch*>(top->successors[1]);
    assert(top_sw);

    if(debug) std::cerr<<"stroot="<<refmap[stroot]<<" boot="<<refmap[boot]<<'\n';

    dfs(top);  if(debug) std::cerr<<"dfs ok\n";
    dfs_black.clear();
    dfs_signals(top, 0, NULL); if(debug) std::cerr<<"dfs_signals 0 ok\n";
    dfs_black.clear();
    dfs_signals(top, 1, &depth_local_signals); if(debug) std::cerr<<"dfs_signals 1 ok\n";

    local_enters_valid[NULL]=w_0;
    fin_loop_valid[NULL]=w_0;
  }

  void GRC2zaphod::setup_sm(STNode *n) // bottom to up
  {

    for(vector<GRCNode*>::iterator ch=n->successors.begin();
	ch!=n->successors.end();ch++) if(*ch){
	  setup_sm((STNode*)*ch);
	}

    if (dynamic_cast<STexcl*>(n)) {
      machines[n] = new statemachine(n, this);
    }

  }
    
  void GRC2zaphod::build_sm(STNode *n, wire *susp)
  {
    if (dynamic_cast<STexcl*>(n)){
      machines[n]->sm2circuit(susp);
    }

    if(suspend.count(n) > 0) {
      susp = current-> build_orgate(susp, suspend[n]);
      new strap(susp,"SUS",refmap[n], 3);
    }
    suspend[n] = susp;

    for(vector<GRCNode*>::iterator ch=n->successors.begin(); ch!=n->successors.end();ch++){
      if(*ch)
	build_sm((STNode*)*ch, susp);
    }  
  }
    

  void GRC2zaphod::execute()
  {
    int tsz,i;
    SignalSymbol *sigsym;
    wire *sigw;
    STleaf *fin;

    int main_thr_t0;

    debug = 0;
    sdebug= 0;
    spdebug=0;
    edebug =0;

    if(action == "vdelays") cv.delay=1; // not working now

    init(); // general info, topolist

    fin=dynamic_cast<STleaf*>(stroot->successors[0]);
    if( fin && fin->isfinal()) main_thr_t0 =1; else main_thr_t0 =0;


    // initialize all signals
    for(SymbolTable::iterator isym = sigs->begin(); isym != sigs->end(); isym++)
      if((sigsym=dynamic_cast<SignalSymbol*>(*isym))){
	  sigw=new wire("Osig"+sigsym->name);
	  sigw->monitor=3;
	  if(sigsym->direction == "output" || sigsym->direction == "inputoutput"){
	    sigw->required=sigw->output = 1;
	  }
	  signals[sigsym].signal=sigw;
      }

    setup_sm(stroot);
    
    tsz=topo.size();
    for(i=tsz-1; i>=0; i--){
      topo[i]->welcome(*this);
    }

    link_syncs();
 
    // close signals
    for(map<SignalSymbol*, zaphod_signal>::iterator sig = signals.begin();
	sig != signals.end(); sig++){
      sigsym = (*sig).first;
      if(sigsym->direction == "input" || sigsym->direction == "inputoutput"){
	if(sigsym->name != "tick"){
	  // add external emitter
	  sigw = new wire("Isig"+sigsym->name);
	  sigw->input = 1;
	  signals[sigsym].sigemit.push_back(sigw);
	  new strap(sigw, w_0); // no real world signals yet
	} else {
	  // tick - discard bogus emitters
	  signals[sigsym].sigemit.clear();
	  signals[sigsym].sigemit.push_back(w_1);}
      }
      sigw=current->build_orgate(signals[sigsym].sigemit);
      new strap(signals[sigsym].signal, sigw?sigw:w_0);
    }
    
    zaphod_signal *zs;
    // link local signals to emmiters
    for(sigview::iterator sig = flat_local_signals.begin();
	sig!=flat_local_signals.end(); sig++){
      sigsym = (*sig).first;
      zs = flat_local_signals[sigsym];
      sigw=current->build_orgate(zs->sigemit);
      new strap(zs->signal, sigw?sigw:w_0);
      new strap(zs->signal, zs->realname, 3);
      std::cerr<<"new signal closing for local "<<zs->realname<<"\n";
    }

    build_sm(stroot, w_0);
    build_counters();

    if(main_thr_t0)
      new strap(w_finished, machines[stroot]->getstate(0));
    else
      new strap(w_finished, w_0); 

    if(debug)std::cerr<<"Circuit in memory OK?\n";
    
    if(action=="verilog"){
      cv.remove_unused();

      monitor_level=2;
      cv.build();
      
      monitor_level=3;
      cv.drawdot();
    }

    if(action=="blif"){
      cv.remove_unused();
      cv.build_blif();
    }


    if(debug)std::cerr<<"Circuit dumped OK\n";
  }

  void GRC2zaphod::children_sameflow(GRCNode *n, wire *flow)
  {
    if(!flow){
      if(sdebug) std::cerr<<"  no flow from"<<refmap[n]<<'\n';
      return;
    }
    for(vector<GRCNode*>::iterator c=n->successors.begin(); c!=n->successors.end();c++)
      if(*c) {
	cflow[n][*c]=flow;
	if(sdebug) std::cerr<<"  set flow "<<refmap[n]<<" -> "<<refmap[*c]<<" to "<<flow->name<<'\n';
      }
  }
  
  wire *GRC2zaphod::flow_pair(GRCNode *p, GRCNode *c)
  {
    if (cflow.count(p) && cflow[p].count(c)) {
      if(cflow[p][c] == w_0) return NULL; else return cflow[p][c];
    }
    return NULL;
  }

  wire *GRC2zaphod::flow_fromparents(GRCNode *n)
  {
    vector <wire*> tmp;
    wire *w;

    if(sdebug)std::cerr<<n->predecessors.size()<<" parent[s] found\n";
    for(vector<GRCNode*>::iterator p=n->predecessors.begin(); p!=n->predecessors.end();p++){
      w = flow_pair(*p, n);
      if(sdebug) std::cerr<<"   flow "<<refmap[*p]<<" -> "<<refmap[n]<<" is "<<current->wn(w)<<'\n';
      tmp.push_back(flow_pair(*p, n));
    }
    w =  current-> build_orgate(tmp);
    return w;
  }

  wire *GRC2zaphod::dotest(GRCNode *n, wire *flowin, wire *wtest) // can have both children not NULL or not
  {
    wire *wntest, *w, *wtrue;
    GRCNode *ch;

    vector<wire*> tmp;

    assert(wtest);

    assert(n->successors.size()==2);

    if(!flowin || flowin == w_0) return NULL;
    ch=n->successors[0];
    if(ch){
      wntest = current-> build_notgate(wtest);
      w = current-> build_andgate(wntest, flowin);
      cflow[n][ch]=w;
      if(sdebug) std::cerr<<"  set flow "<<refmap[n]<<" -> "<<refmap[ch]<<" to "<<current->wn(w)<<'\n';
    }
    ch=n->successors[1];
    if(ch){
      w = current-> build_andgate(wtest, flowin);
      cflow[n][ch]=w;
      if(sdebug) std::cerr<<"  set flow "<<refmap[n]<<" -> "<<refmap[ch]<<" to "<<current->wn(w)<<'\n';
      wtrue = w;
    } else wtrue = NULL;
    return wtrue;
  }


  // SWITCH
  
  Status GRC2zaphod::visit(Switch &s)
  {

    wire* flowin;
    wire *chk, *flowch;
    GRCNode *ch;
    int i, sz;

    if(debug)std::cerr<<"switch "<<refmap[&s]<<'\n';
    flowin = flow_fromparents(&s);
    
    cflow[&s][NULL] = flowin; // it's own flow

    if ((dynamic_cast<STexcl*> (s.st)) == NULL) {
      std::cerr<<"switch doesn't point to STexcl\n";
      assert(0);
      return Status();
    }
    
    sz=s.successors.size();
    for(i=0;i<sz;i++){
      ch=s.successors[i];
      if (ch){
	chk = machines[s.st]->getstate(i);
	flowch=current->build_andgate(chk, flowin);

	if(flowch) {
	  if(flow_pair(&s,ch))
	    flowch = current->build_orgate(flowch, cflow[&s][ch]);
	  cflow[&s][ch] = flowch;
	}
	if(sdebug) std::cerr<<"  set flow "<<refmap[&s]<<" -> "<<refmap[ch]<<" to "<<current->wn(flowch)<<'\n';
      }
    }
    
    return Status();
    
  }

  // ENTER and LEAVE
  
  Status GRC2zaphod::visit(Enter &s)
  {
    STNode *par;
    STexcl *ex;
    wire *flowin;

    int i, ch, sz;

    if(debug) std::cerr<<"Enter "<<refmap[&s]<<'\n';

    flowin=flow_fromparents(&s);

    children_sameflow(&s, flowin);

    par=(STNode*)s.st->predecessors[0];
    
    if((ex=dynamic_cast<STexcl*>(par))==NULL) return Status(); //father is not exclusive node

    // look to parent; which child am I ?
    ch=-1;
    sz=par->successors.size();
    for(i=0;i<sz;i++)
      if(s.st == par->successors[i]) {ch=i; break;}

    assert(ch>=0);

    machines[par]->addenter(ch,flowin, next_sync[&s]);

    return Status();
  }

  Status GRC2zaphod::visit(Leave &s)
  {
    wire *flowin;
    if(debug) std::cerr<<"Leave "<<refmap[&s]<<'\n';
    flowin=flow_fromparents(&s);
    children_sameflow(&s, flowin);
    return Status();
  }

  // ACTION (emit signal) and TEST (check signal)

  Status GRC2zaphod::visit(Action &s)
  {
    wire *flowin;

    if(debug)std::cerr<<"action "<<refmap[&s]<<'\n';
    flowin=flow_fromparents(&s);

    zc.local_sig = localsignals[&s];
    zc.do_statement(s.body, flowin);

    children_sameflow(&s, flowin);

    return Status();
    
  }

  Status GRC2zaphod::visit(Test &s)
  {
    wire *sp, *flowin;
    STref *ref;

    if(debug) std::cerr<<"Test  "<<refmap[&s]<<'\n';

    flowin=flow_fromparents(&s);

    zc.local_sig = localsignals[&s];
    sp = dotest(&s, flowin, zc.getbool(s.predicate));
   
    if((ref = dynamic_cast<STref*>(s.st)))
      if(ref && ref->issuspend()){
	if(spdebug)std::cerr<<" sp activator "<<current->wn(sp)<<"\n";
	suspend[s.st] = sp;
      }

    return Status(); 
  }

  // TERMINATE, SYNC and FORK

  Status GRC2zaphod::visit(Terminate &s)
  {
    wire *flowin;
    if(debug)    std::cerr<<"term "<<refmap[&s]<<'\n';    
    if(s.code != 1) {
      flowin=flow_fromparents(&s);
      children_sameflow(&s, flowin);
    }
    return Status();
  }

  Status GRC2zaphod::visit(Sync &s) // the coolest part
  {

    map<int, wire*> flowin;
    map<int, wire*> notflowin;
    vector<GRCNode*> &children = s.successors;

    int i, j, sz,maxcode, mincode, novalid;
    int nothreads, no_t0, can_t0;


    Terminate * t;
    wire *w;

    vector<wire*> tmp;

    if(debug)    std::cerr<<"sync "<<refmap[&s]<<'\n';

    assert(dynamic_cast<STpar*>(s.st));
    nothreads = s.st->successors.size();

    maxcode = 0; mincode=1000; novalid=0; // mincode includes 0 only if can_t0
    no_t0=0;

    sz = s.predecessors.size();
    if(sdebug)std::cerr<<sz<<" parent[s] found\n";
    for (i = 0; i < sz; i++){
      assert(t = dynamic_cast<Terminate*>(s.predecessors[i]));
      w = flow_pair(t, &s);
      if(t->code==1) continue;
      if (w && t->code>maxcode) maxcode = t->code;
      if (w && t->code==0) { no_t0++; continue; }
      if (w && t->code<mincode) mincode = t->code;
      if (w) novalid++;

      if(w) flowin[t->code] = w;
      assert((int)children.size() > t->code);
    }

    can_t0 = (nothreads == no_t0);
    if(debug) std::cerr<<"nothr="<<nothreads<<" no_t0="<<no_t0<<"  upvalid="<<novalid<<'\n';

    // now we know if parallel can terminate 0 and also have all in flows, except 1

    if(can_t0){ // build the AND net for terminate 0
      mincode=0;
      novalid++;
      assert(no_t0 >= 2);
      tmp.clear();
      for (i = 0; i < sz; i++){
	t = dynamic_cast<Terminate*>(s.predecessors[i]);
	if(t->code != 0) continue;
	assert(cflow[t][&s]);
	tmp.push_back(cflow[t][&s]);
      }
      flowin[0] = current-> build_andgate(tmp);

    }
    
    if(debug) std::cerr<<"mincode="<<mincode<<" maxcode="<<maxcode<<" novalid="<<novalid<<'\n';

    // build the gates for all levels except 1, which has no flowin
    // I'll do it flash, as propagation time is essential here

    for(i=maxcode;i>=mincode;i--){
      if(flowin.count(i)){ // have flowin for level i
	//std::cerr<<"have flow on parent"<<i<<"\n";
	assert(children[i]);
	if(i!=mincode) {notflowin[i] = current-> build_notgate(flowin[i]);}
	if(i==maxcode) w=flowin[i]; else {
	  tmp.clear();
	  tmp.push_back(flowin[i]);
	  //	  std::cerr<<"    push direct"<<current->wn(flowin[i])<<"\n";
	  for(j=i+1; j<=maxcode; j++)
	    if(notflowin.count(j)){
	      tmp.push_back(notflowin[j]);
	      //std::cerr<<"    push inv"<<current->wn(notflowin[j])<<"\n";
	    }
	  w = current-> build_andgate(tmp);
	}
	cflow[&s][children[i]]=w;
	//     std::cerr<<"building sync "<<i<<" is "<<current->wn(w)<<"\n";
      }
    }

    // build the local_enters_invalid and fin_loop_invalid wires

    // local_enters_invalid if the term level > 1
    // fin_loop_invalid if above or t0

    tmp.clear();
    for(i=2; i<=maxcode; i++)
      if(flowin.count(i)) tmp.push_back(flowin[i]);
    w = current->build_orgate(tmp);
    local_enters_valid[&s] = w; // must be negated in link_syncs after or !

    new strap(w,"U",refmap[&s],2);

    if(can_t0){
      w=flowin[0];
    } else w=NULL;
    fin_loop_valid[&s] = w; // just t0, must be ored w/ higher levels and negated !
    
    new strap(w,"T",refmap[&s],2);

    return Status();
  }

  Status GRC2zaphod::visit(Fork &s)
  {
    wire *flowin;

    if(debug)    std::cerr<<"fork "<<refmap[&s]<<'\n';

    flowin=flow_fromparents(&s);
    children_sameflow(&s, flowin);

    return Status();
  }

  Status GRC2zaphod::visit(Nop &s)
  {
    wire *flowin;
    GRCNode *ch0, *ch1, *par;

    if(debug)    std::cerr<<"Nop "<<refmap[&s]<<'\n';
    flowin=flow_fromparents(&s);

    if(s.isshortcut()){
      assert(s.successors.size() == 2);
      assert(s.predecessors.size() == 1);
      ch0 = s.successors[0];
      ch1 = s.successors[1];
      par = s.predecessors[0];

      cflow[&s][ch0] = cflow[par][NULL];
      if(sdebug) std::cerr<<"  set flow "<<refmap[&s]<<" -> "<<refmap[ch0]<<" to "<<current->wn(cflow[&s][ch0])<<'\n';

      cflow[&s][ch1] = flowin;
      if(sdebug) std::cerr<<"  set flow "<<refmap[&s]<<" -> "<<refmap[ch1]<<" to "<<current->wn(cflow[&s][ch1])<<'\n';
      
    } else {
      children_sameflow(&s, flowin);
    }
    return Status();
  }


  // RECONSTRUCT T1
  
  Status GRC2zaphod::visit(RecT1 &s) // this is also cool: reconstructs T1
  {

    Nop *nop;
    wire *all_flow, *flowout, *w, *wn;
    vector<wire*> flow_nop, flow_in;

    for(vector<GRCNode*>::iterator p = s.predecessors.begin(); p != s.predecessors.end(); p++){
      nop = dynamic_cast<Nop*>(*p);
      assert(nop);
      w=flow_pair(nop, &s);
      if(nop->isflowin()) flow_in.push_back(w); else flow_nop.push_back(w);
    }
    w = current-> build_orgate(flow_nop);
    wn = current-> build_notgate(w);
    all_flow = current-> build_orgate(flow_in);
    flowout = current-> build_andgate(wn, all_flow);

    children_sameflow(&s, flowout);

    new strap(flowout, "RT",refmap[&s], 3);

    return Status();
  }


  // DEFINE SIGNAL

  Status GRC2zaphod::visit(DefineSignal &s)
  {
    // wire *new_sig;
    //    SignalSymbol *ss;
    wire* flowin;

    if(debug)    std::cerr<<"DefineSignal "<<refmap[&s]<<'\n';
    
    flowin=flow_fromparents(&s);
    children_sameflow(&s, flowin);

    return Status();
    /*
    for(SymbolTable::iterator isym = sigs->begin(); isym != sigs->end(); isym++){
      assert((ss=dynamic_cast<SignalSymbol*>(*isym)));

      if(signals.count(ss) > 0)
	new strap(signals[ss], current->build_orgate(sigemit[ss]));

      new_sig = new wire(ss->name, refmap[&s]);
      signals[ss] = new_sig;
      sigemit[ss].clear();
    }
    return Status();
    */
  }

  Status GRC2zaphod::visit(TopGRC &s)
  {
    children_sameflow(&s,w_1);
    return Status();
  }

}

int main(int argc, char *argv[])
{

  if(argc<6) {std::cerr<<"Usage: ... action grc_file dot_file vfile vdotfile\n"; return -1;}

  string action(argv[1]);

  std::ifstream grcfile(argv[2]); if(!grcfile.good()) return -1;
  std::ofstream dotfile(argv[3]); if(!dotfile.good()) return -1;
  std::ofstream vfile(argv[4]); if(!vfile.good()) return -1;
  std::ofstream vdotfile(argv[5]); if(!vdotfile.good()) return -1;

  try {

    IR::XMListream r(grcfile);
    IR::Node *n;
    r >> n;
    grcfile.close();

    AST::Modules *mods = dynamic_cast<AST::Modules*>(n);
    if (!mods) throw IR::Error("Root node is not a Modules object");

    for ( std::vector<AST::Module*>::iterator i = mods->modules.begin() ;
	  i != mods->modules.end() ; i++ ) {
      assert(*i);

      AST::TopGRC *gn = dynamic_cast<AST::TopGRC*>((*i)->body);
      assert(gn);
      AST::SymbolTable *sig = dynamic_cast<AST::SymbolTable*>((*i)->signals);
      assert(sig);

      if(0) std::cerr << "File read - OK"<<std::endl;

      AST::Nodemap refmap;
      AST::GRCDot(dotfile,(AST::GRCNode *)gn,(*i),refmap);
      dotfile.close();


      ZAPHOD::GRC2zaphod tw (vfile, vdotfile, *i, refmap, action);

      tw.execute();
      
      vfile.close(); vdotfile.close();
    }

  } catch (IR::Error &e) {
    std::cerr << e.s << std::endl;
    exit(-1);
  }

  return 0;

}
